from http import HTTPStatus
from unittest.mock import AsyncMock, patch

from synapse.api.errors import Codes, SynapseError
from synapse.server import HomeServer
from synapse.util.clock import Clock
from twisted.internet.testing import MemoryReactor

from famedly_control_synapse.rest.types import CreateManagedRoomRequest
from famedly_control_synapse.room_handler import famedly_control_user_sync_error
from tests.utils.module_api_testcase import ModuleApiTestCase


class TestRoomHandler(ModuleApiTestCase):

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.room_handler = self.hs.room_control.room_handler
        self.server_name = self.hs.hostname

    def _check_users_joined_to_room(
        self,
        room: str,
        members: list[str],
    ) -> None:
        """Helper method to check if the users have joined a room."""
        for member in members:
            path = "/rooms/%s/state/m.room.member/%s" % (room, member)
            channel = self.make_request(
                "GET", path, access_token=self.creator_access_token
            )
            assert channel.code == HTTPStatus.OK, channel.result
            assert channel.json_body["membership"] == "join", channel.result

    def _create_managed_room_without_members(
        self,
    ) -> str:
        config = CreateManagedRoomRequest(
            room_alias_name="membership_test_room",
            name="Membership Test Room",
            groups=["test_group"],
        )
        with (
            patch(
                "famedly_control_synapse.client.FamedlyControlClient.get_group_members",
                new_callable=AsyncMock,
                return_value=[],
            ),
            patch(
                "famedly_control_synapse.room_handler.ManagedRoomHandler.batch_convert_external_user_ids_to_matrix_user_ids",
                new_callable=AsyncMock,
                side_effect=lambda x: (x, []),
            ),
        ):
            channel = self.make_request(
                method="POST",
                path=self.CREATE_PATH,
                content=config.model_dump(),
                access_token=self.creator_access_token,
                shorthand=False,
            )
        assert channel.code == HTTPStatus.OK, channel.result
        return channel.json_body["room_id"]

    def test_force_join_users_to_room(self) -> None:
        """Test that users can be force joined to a room."""
        user_a = self.register_user("test_user_a", "password")
        user_b = self.register_user("test_user_b", "password")
        user_c = self.register_user("test_user_c", "password")
        room_id = self._create_managed_room_without_members()

        error = self.get_success(
            self.room_handler.force_join_users_to_room(
                room_id=room_id,
                user_mxids=[user_a, user_b, user_c],
                requester=self.requester,
            )
        )
        assert error is None, f"Expected no errors, but got: {error}"
        self._check_users_joined_to_room(room_id, [user_a, user_b, user_c])

    def test_force_join_users_already_in_room(self) -> None:
        """Test that force joining users who are already in the room is handled correctly."""
        user_a = self.register_user("test_user_a", "password")
        user_b = self.register_user("test_user_b", "password")
        room_id = self._create_managed_room_without_members()

        # Force join the same user twice
        error = self.get_success(
            self.room_handler.force_join_users_to_room(
                room_id=room_id, user_mxids=[user_a, user_b], requester=self.requester
            )
        )
        assert error is None, f"Expected no errors, but got: {error}"
        self._check_users_joined_to_room(room_id, [user_a, user_b])

        error = self.get_success(
            self.room_handler.force_join_users_to_room(
                room_id=room_id, user_mxids=[user_a, user_b], requester=self.requester
            )
        )
        assert error is None, f"Expected no errors, but got: {error}"
        self._check_users_joined_to_room(room_id, [user_a, user_b])

    def test_force_join_users_including_failed_users(self) -> None:
        """Test that force joining users handles failures in a mixed scenario."""
        # Get the metric initial value
        initial_value = famedly_control_user_sync_error.labels(
            error_code="M_BAD_STATE", server_name=self.server_name
        )._value.get()

        user_a = self.register_user("test_user_a", "password")
        user_b = self.register_user("test_user_b", "password")
        room_id = self._create_managed_room_without_members()

        # Ban user_b to make force join fail for this user
        self.helper.ban(
            room_id, self.creator, user_b, HTTPStatus.OK, self.creator_access_token
        )

        # Attempt to force join a mix of users where one will fail (banned user)
        error = self.get_success(
            self.room_handler.force_join_users_to_room(
                room_id=room_id, user_mxids=[user_a, user_b], requester=self.requester
            )
        )
        # Should return an error dict with user_b (the banned user)
        assert error is not None, "Expected errors, but got None"
        assert user_b in error, f"Expected error for {user_b}, but got: {error}"
        # Check that the valid user (user_a) successfully joined the room
        self._check_users_joined_to_room(room_id, [user_a])

        # Check the metric value is increased by 1
        new_value = famedly_control_user_sync_error.labels(
            error_code="M_BAD_STATE", server_name=self.server_name
        )._value.get()
        assert new_value == initial_value + 1

    def test_force_join_users_with_nonexistent_user(self) -> None:
        """Test that force_join_users_to_room catches error for non-existent user."""
        # Get the metric initial value
        initial_value = famedly_control_user_sync_error.labels(
            error_code="M_NOT_FOUND", server_name=self.server_name
        )._value.get()

        user_a = self.register_user("test_user_a", "password")
        non_existent_user = f"@random_user:{self.server_name}"
        room_id = self._create_managed_room_without_members()

        error = self.get_success(
            self.room_handler.force_join_users_to_room(
                room_id=room_id,
                user_mxids=[non_existent_user, user_a],
                requester=self.requester,
            )
        )
        # Should return an error dict with the non-existent user
        assert error is not None, "Expected errors, but got None"
        assert (
            non_existent_user in error
        ), f"Expected error for {non_existent_user}, but got: {error}"
        assert error[non_existent_user] == "User does not exist on this server"

        # Check that the valid user (user_a) successfully joined the room
        self._check_users_joined_to_room(room_id, [user_a])

        # Check the metric value is increased by 1
        new_value = famedly_control_user_sync_error.labels(
            error_code="M_NOT_FOUND", server_name=self.server_name
        )._value.get()
        assert new_value == initial_value + 1

    def test_remove_users_from_room(self) -> None:
        """Test that users can be removed from a room."""
        user_a = self.register_user("test_user_a", "password")
        user_b = self.register_user("test_user_b", "password")
        room_id = self._create_managed_room_without_members()

        # Add the users to the room to make sure they can be removed
        error = self.get_success(
            self.room_handler.force_join_users_to_room(
                room_id=room_id, user_mxids=[user_a, user_b], requester=self.requester
            )
        )
        assert error is None, f"Expected no errors, but got: {error}"
        self._check_users_joined_to_room(room_id, [user_a, user_b])

        # Now remove the users from the room
        error = self.get_success(
            self.room_handler.remove_users_from_room(
                creator_id=self.creator, user_mxids=[user_a, user_b], room_id=room_id
            )
        )
        assert error is None, f"Expected no errors, but got: {error}"
        # Check that the users have been removed
        for member in [user_a, user_b]:
            path = "/rooms/%s/state/m.room.member/%s" % (room_id, member)
            channel = self.make_request(
                "GET", path, access_token=self.creator_access_token
            )
            assert channel.code == HTTPStatus.OK, channel.result
            assert channel.json_body["membership"] == "leave", channel.result

    def test_remove_users_from_room_with_nonexistent_user(self) -> None:
        """Test that remove_users_from_room catches user non-existent case."""
        # Get the metric initial value
        initial_value = famedly_control_user_sync_error.labels(
            error_code="M_NOT_FOUND", server_name=self.server_name
        )._value.get()

        user_a = self.register_user("test_user_a", "password")
        non_existent_user = f"@nonexistent:{self.server_name}"
        room_id = self._create_managed_room_without_members()

        # Add the valid user to the room
        error = self.get_success(
            self.room_handler.force_join_users_to_room(
                room_id=room_id, user_mxids=[user_a], requester=self.requester
            )
        )
        assert error is None, f"Expected no errors, but got: {error}"
        self._check_users_joined_to_room(room_id, [user_a])

        # Now attempt to remove both the valid user and the non-existent user
        error = self.get_success(
            self.room_handler.remove_users_from_room(
                creator_id=self.creator,
                user_mxids=[user_a, non_existent_user],
                room_id=room_id,
            )
        )
        # Should return an error dict with the non-existent user
        assert error is not None, "Expected errors, but got None"
        assert (
            non_existent_user in error
        ), f"Expected error for {non_existent_user}, but got: {error}"
        # Check that the valid user (user_a) has been removed from the room
        path = "/rooms/%s/state/m.room.member/%s" % (room_id, user_a)
        channel = self.make_request("GET", path, access_token=self.creator_access_token)
        assert channel.code == HTTPStatus.OK, channel.result
        assert channel.json_body["membership"] == "leave", channel.result

        # Check the metric value is increased by 1
        new_value = famedly_control_user_sync_error.labels(
            error_code="M_NOT_FOUND", server_name=self.server_name
        )._value.get()
        assert new_value == initial_value + 1

    def test_remove_users_from_room_with_not_a_member_user(self) -> None:
        """Test that remove_users_from_room skips the user is not a member case."""
        user_a = self.register_user("test_user_a", "password")
        non_member_user = self.register_user("non_member_user", "password")
        room_id = self._create_managed_room_without_members()

        # Add a valid user to the room
        error = self.get_success(
            self.room_handler.force_join_users_to_room(
                room_id=room_id, user_mxids=[user_a], requester=self.requester
            )
        )
        assert error is None, f"Expected no errors, but got: {error}"
        self._check_users_joined_to_room(room_id, [user_a])

        # Now attempt to remove both the valid user and the non-member user
        error = self.get_success(
            self.room_handler.remove_users_from_room(
                creator_id=self.creator,
                user_mxids=[user_a, non_member_user],
                room_id=room_id,
            )
        )
        # This does not return an error because the method should skip the non-member user and still remove the valid user
        assert error is None
        # Check that the valid user (user_a) has been removed from the room
        path = "/rooms/%s/state/m.room.member/%s" % (room_id, user_a)
        channel = self.make_request("GET", path, access_token=self.creator_access_token)
        assert channel.code == HTTPStatus.OK, channel.result
        assert channel.json_body["membership"] == "leave", channel.result

    def test_remove_users_unexpected_error(self) -> None:
        # Get the metric initial value
        initial_value = famedly_control_user_sync_error.labels(
            error_code="M_UNKNOWN", server_name=self.server_name
        )._value.get()

        user_a = self.register_user("test_user_a", "password")
        room_id = self._create_managed_room_without_members()

        # Add the valid user to the room
        error = self.get_success(
            self.room_handler.force_join_users_to_room(
                room_id=room_id, user_mxids=[user_a], requester=self.requester
            )
        )
        assert error is None, f"Expected no errors, but got: {error}"
        self._check_users_joined_to_room(room_id, [user_a])

        # Now attempt to remove the valid user and simulate an unexpected error
        with patch.object(
            self.room_handler.api, "update_room_membership", new_callable=AsyncMock
        ) as mock_update:
            mock_update.side_effect = SynapseError(
                500, "Unexpected error", Codes.UNKNOWN
            )

            error = self.get_success(
                self.room_handler.remove_users_from_room(
                    creator_id=self.creator,
                    user_mxids=[user_a],
                    room_id=room_id,
                )
            )
            assert error is not None, "Expected errors, but got None"
            assert user_a in error, f"Expected error for {user_a}, but got: {error}"

        # Check that the the user has not been removed from the room due to the unexpected error
        path = "/rooms/%s/state/m.room.member/%s" % (room_id, user_a)
        channel = self.make_request("GET", path, access_token=self.creator_access_token)
        assert channel.code == HTTPStatus.OK, channel.result
        assert channel.json_body["membership"] == "join", channel.result

        # Check the metric value is increased by 1
        new_value = famedly_control_user_sync_error.labels(
            error_code="M_UNKNOWN", server_name=self.server_name
        )._value.get()
        assert new_value == initial_value + 1

    def test_batch_convert_external_user_ids_to_matrix_user_ids(self) -> None:
        """Test that batch conversion of external user IDs to Matrix user IDs works correctly."""
        external_user_ids = []
        matrix_user_ids = []

        # Generate 105 pairs of matrix_user_id and external_user_id
        for i in range(105):
            external_user_id = f"external_user_{i}"
            external_user_ids.append(external_user_id)
            matrix_user_id = f"@matrix_user_{i}:{self.server_name_for_this_server}"
            matrix_user_ids.append(matrix_user_id)

        # Insert into the user_external_ids table
        self.get_success(
            self.store.db_pool.simple_insert_many(
                table="user_external_ids",
                keys=["user_id", "external_id", "auth_provider"],
                values=[
                    (matrix_user_id, external_id, "https://idp.example.com/")
                    for matrix_user_id, external_id in zip(
                        matrix_user_ids, external_user_ids
                    )
                ],
                desc="test_insert_multiple_user_external_ids",
            )
        )

        # Test the batch conversion method
        result, not_found = self.get_success(
            self.room_handler.batch_convert_external_user_ids_to_matrix_user_ids(
                external_user_ids
            )
        )
        assert result == matrix_user_ids
        assert not_found == []
        assert len(result) == 105

    def test_batch_convert_with_missing_users(self) -> None:
        """Test that batch conversion handles missing users gracefully.

        This can happen when users exist in Zitadel but haven't logged into Synapse yet.
        The method should return partial results instead of raising an exception.
        """
        # Mix of existing and non-existing users
        external_user_ids = [
            "existing_user_1",
            "missing_user_1",  # This user is in Zitadel but not in Synapse
            "existing_user_2",
            "missing_user_2",  # This user is in Zitadel but not in Synapse
        ]

        # Only insert the "existing" users
        self.get_success(
            self.store.db_pool.simple_insert_many(
                table="user_external_ids",
                keys=["user_id", "external_id", "auth_provider"],
                values=[
                    (
                        f"@existing_1:{self.server_name_for_this_server}",
                        "existing_user_1",
                        "https://idp.example.com/",
                    ),
                    (
                        f"@existing_2:{self.server_name_for_this_server}",
                        "existing_user_2",
                        "https://idp.example.com/",
                    ),
                ],
                desc="test_insert_partial_user_external_ids",
            )
        )

        result, not_founds = self.get_success(
            self.room_handler.batch_convert_external_user_ids_to_matrix_user_ids(
                external_user_ids
            )
        )

        assert len(result) == 2
        assert len(not_founds) == 2
        assert f"@existing_1:{self.server_name_for_this_server}" in result
        assert f"@existing_2:{self.server_name_for_this_server}" in result
        assert "missing_user_1" in not_founds
        assert "missing_user_2" in not_founds

    @patch(
        "famedly_control_synapse.client.FamedlyControlClient.get_group_members",
        new_callable=AsyncMock,
    )
    @patch(
        "famedly_control_synapse.room_handler.ManagedRoomHandler.batch_convert_external_user_ids_to_matrix_user_ids",
        new_callable=AsyncMock,
    )
    def test_get_room_creator(self, mock_batch_convert, mock_get_group_members) -> None:
        # Create a room and get its ID
        group_1 = "group_1"
        member_a = self.register_user("test_member_a", "password")
        member_b = self.register_user("test_member_b", "password")
        group_1_members = [member_a, member_b]

        mock_get_group_members.return_value = (
            group_1_members  # in real case this should be external_ids
        )
        mock_batch_convert.side_effect = lambda x: (x, [])

        room_id = self._create_managed_room(
            name="Test Room with Group Members", groups=[group_1]
        )
        # Get the room creator using the method
        creator = self.get_success(self.room_handler.get_room_creator(room_id))

        # Assert that the creator is correct
        assert creator == self.creator



from http import HTTPStatus
from unittest.mock import AsyncMock, patch

from synapse.server import HomeServer
from synapse.util.clock import Clock
from twisted.internet.testing import MemoryReactor

from famedly_control_synapse.rest.types import CreateManagedRoomRequest
from tests.utils.module_api_testcase import ModuleApiTestCase


class TestMembershipHook(ModuleApiTestCase):
    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.non_admin = self.register_user("non_admin", "password", admin=False)
        self.non_admin_token = self.login("non_admin", "password")

    def _create_managed_room(
        self, name: str = "Membership Test Room", groups: list[str] | None = None
    ) -> str:
        config = CreateManagedRoomRequest(
            room_alias_name="membership_test_room",
            name=name,
            groups=groups or ["test_group"],
        )
        with (
            patch(
                "famedly_control_synapse.client.FamedlyControlClient.get_group_members",
                new_callable=AsyncMock,
                return_value=[],
            ),
            patch(
                "famedly_control_synapse.room_handler.ManagedRoomHandler.batch_convert_external_user_ids_to_matrix_user_ids",
                new_callable=AsyncMock,
                side_effect=lambda x: (x, []),
            ),
        ):
            channel = self.make_request(
                method="POST",
                path=self.CREATE_PATH,
                content=config.model_dump(),
                access_token=self.creator_access_token,
                shorthand=False,
            )
        assert channel.code == HTTPStatus.OK, channel.result
        return channel.json_body["room_id"]

    def _create_managed_room_with_non_admin(self) -> str:
        """Create a managed room and add the non-admin user to it."""
        room_id = self._create_managed_room()
        self.helper.invite(
            room_id,
            src=self.creator,
            targ=self.non_admin,
            tok=self.creator_access_token,
        )
        self.helper.join(room_id, user=self.non_admin, tok=self.non_admin_token)
        return room_id

    def test_admin_can_invite(self) -> None:
        """The admin should be able to invite users to a managed room."""
        room_id = self._create_managed_room()
        self.helper.invite(
            room_id,
            src=self.creator,
            targ=self.non_admin,
            tok=self.creator_access_token,
            expect_code=HTTPStatus.OK,
        )

    def test_non_admin_cannot_invite(self) -> None:
        """A non-admin user should not be able to invite others to a managed room."""
        room_id = self._create_managed_room_with_non_admin()
        other_user = self.register_user("other_user", "password", admin=False)
        self.helper.invite(
            room_id,
            src=self.non_admin,
            targ=other_user,
            tok=self.non_admin_token,
            expect_code=HTTPStatus.FORBIDDEN,
        )

    def test_non_admin_cannot_leave(self) -> None:
        """A non-admin user should not be able to leave a managed room."""
        room_id = self._create_managed_room_with_non_admin()
        self.helper.leave(
            room_id,
            user=self.non_admin,
            tok=self.non_admin_token,
            expect_code=HTTPStatus.FORBIDDEN,
        )

    def test_admin_can_ban(self) -> None:
        """The admin should be able to ban users from a managed room."""
        room_id = self._create_managed_room_with_non_admin()
        self.helper.ban(
            room_id,
            src=self.creator,
            targ=self.non_admin,
            tok=self.creator_access_token,
            expect_code=HTTPStatus.OK,
        )

    def test_non_admin_cannot_ban(self) -> None:
        """A non-admin user should not be able to ban others in a managed room."""
        room_id = self._create_managed_room_with_non_admin()
        other_user = self.register_user("ban_target", "password", admin=False)
        # Admin adds the other user first
        self.helper.invite(
            room_id,
            src=self.creator,
            targ=other_user,
            tok=self.creator_access_token,
        )
        # Non-admin tries to ban
        self.helper.ban(
            room_id,
            src=self.non_admin,
            targ=other_user,
            tok=self.non_admin_token,
            expect_code=HTTPStatus.FORBIDDEN,
        )

    def test_admin_can_kick(self) -> None:
        """The admin should be able to kick users from a managed room."""
        room_id = self._create_managed_room_with_non_admin()
        self.helper.change_membership(
            room_id,
            src=self.creator,
            targ=self.non_admin,
            tok=self.creator_access_token,
            membership="leave",
            expect_code=HTTPStatus.OK,
        )

    def test_non_invited_user_cannot_join(self) -> None:
        """A user who has not been invited should not be able to join a managed room."""
        room_id = self._create_managed_room()
        self.helper.join(
            room_id,
            user=self.non_admin,
            tok=self.non_admin_token,
            expect_code=HTTPStatus.FORBIDDEN,
        )

    def test_non_admin_cannot_kick(self) -> None:
        """A non-admin user should not be able to kick others in a managed room."""
        room_id = self._create_managed_room_with_non_admin()
        other_user = self.register_user("kick_target", "password", admin=False)
        # Admin adds the other user first
        self.helper.invite(
            room_id,
            src=self.creator,
            targ=other_user,
            tok=self.creator_access_token,
        )
        # Non-admin tries to kick
        self.helper.change_membership(
            room_id,
            src=self.non_admin,
            targ=other_user,
            tok=self.non_admin_token,
            membership="leave",
            expect_code=HTTPStatus.FORBIDDEN,
        )

    def test_non_managed_room_not_affected(self) -> None:
        """Membership changes in non-managed rooms should not be restricted."""
        room_id = self.helper.create_room_as(
            self.creator, tok=self.creator_access_token, is_public=False
        )
        self.helper.invite(
            room_id,
            src=self.creator,
            targ=self.non_admin,
            tok=self.creator_access_token,
        )
        self.helper.join(
            room_id,
            user=self.non_admin,
            tok=self.non_admin_token,
            expect_code=HTTPStatus.OK,
        )
        self.helper.leave(
            room_id,
            user=self.non_admin,
            tok=self.non_admin_token,
            expect_code=HTTPStatus.OK,
        )



# Copyright 2014-2016 OpenMarket Ltd
# Copyright 2018-2019 New Vector Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import atexit
import os
import sys
import warnings
from collections.abc import Callable
from typing import TYPE_CHECKING, Literal, TypeVar, overload

from synapse.config.homeserver import HomeServerConfig
from synapse.config.server import DEFAULT_ROOM_VERSION
from synapse.storage.database import LoggingDatabaseConnection
from synapse.storage.engines import create_engine
from synapse.storage.prepare_database import prepare_database

if TYPE_CHECKING:
    from sys import UnraisableHookArgs


# set this to True to run the tests against postgres instead of sqlite.
#
# When running under postgres, we first create a base database with the name
# POSTGRES_BASE_DB and update it to the current schema. Then, for each test case, we
# create another unique database, using the base database as a template.
USE_POSTGRES_FOR_TESTS = os.environ.get("SYNAPSE_POSTGRES", False)
LEAVE_DB = os.environ.get("SYNAPSE_LEAVE_DB", False)
POSTGRES_USER = os.environ.get("SYNAPSE_POSTGRES_USER", None)
POSTGRES_HOST = os.environ.get("SYNAPSE_POSTGRES_HOST", None)
POSTGRES_PASSWORD = os.environ.get("SYNAPSE_POSTGRES_PASSWORD", None)
POSTGRES_PORT = (
    int(os.environ["SYNAPSE_POSTGRES_PORT"])
    if "SYNAPSE_POSTGRES_PORT" in os.environ
    else None
)
POSTGRES_BASE_DB = f"_synapse_unit_tests_base_{os.getpid()}"

# When debugging a specific test, it's occasionally useful to write the
# DB to disk and query it with the sqlite CLI.
SQLITE_PERSIST_DB = os.environ.get("SYNAPSE_TEST_PERSIST_SQLITE_DB") is not None

# the dbname we will connect to in order to create the base database.
POSTGRES_DBNAME_FOR_INITIAL_CREATE = "postgres"


def setupdb() -> None:
    # If we're using PostgreSQL, set up the db once
    if USE_POSTGRES_FOR_TESTS:
        # create a PostgresEngine
        db_engine = create_engine({"name": "psycopg2", "args": {}})
        # connect to postgres to create the base database.
        db_conn = db_engine.module.connect(
            user=POSTGRES_USER,
            host=POSTGRES_HOST,
            port=POSTGRES_PORT,
            password=POSTGRES_PASSWORD,
            dbname=POSTGRES_DBNAME_FOR_INITIAL_CREATE,
        )
        db_engine.attempt_to_set_autocommit(db_conn, autocommit=True)
        cur = db_conn.cursor()
        cur.execute(f"DROP DATABASE IF EXISTS {POSTGRES_BASE_DB};")
        cur.execute(
            f"CREATE DATABASE {POSTGRES_BASE_DB} ENCODING 'UTF8' LC_COLLATE='C' LC_CTYPE='C' "
            f"template=template0;"
        )
        cur.close()
        db_conn.close()

        # Set up in the db
        db_conn = db_engine.module.connect(
            dbname=POSTGRES_BASE_DB,
            user=POSTGRES_USER,
            host=POSTGRES_HOST,
            port=POSTGRES_PORT,
            password=POSTGRES_PASSWORD,
        )
        logging_conn = LoggingDatabaseConnection(
            conn=db_conn,
            engine=db_engine,
            default_txn_name="tests",
            server_name="test_server",
        )
        prepare_database(logging_conn, db_engine, None)
        logging_conn.close()

        def _cleanup() -> None:
            db_conn = db_engine.module.connect(
                user=POSTGRES_USER,
                host=POSTGRES_HOST,
                port=POSTGRES_PORT,
                password=POSTGRES_PASSWORD,
                dbname=POSTGRES_DBNAME_FOR_INITIAL_CREATE,
            )
            db_engine.attempt_to_set_autocommit(db_conn, autocommit=True)
            cur = db_conn.cursor()
            cur.execute(f"DROP DATABASE IF EXISTS {POSTGRES_BASE_DB};")
            cur.close()
            db_conn.close()

        atexit.register(_cleanup)


@overload
def default_config(name: str, parse: Literal[False] = ...) -> dict[str, object]: ...


@overload
def default_config(name: str, parse: Literal[True]) -> HomeServerConfig: ...


def default_config(
    name: str, parse: bool = False
) -> dict[str, object] | HomeServerConfig:
    """
    Create a reasonable test config.
    """
    config_dict = {
        "server_name": name,
        # Setting this to an empty list turns off federation sending.
        "federation_sender_instances": [],
        "media_store_path": "media",
        # the test signing key is just an arbitrary ed25519 key to keep the config
        # parser happy
        "signing_key": "ed25519 a_lPym qvioDNmfExFBRPgdTU+wtFYKq4JfwFRv7sYVgWvmgJg",
        # Disable trusted key servers, otherwise unit tests might try to actually
        # reach out to matrix.org.
        "trusted_key_servers": [],
        "event_cache_size": 1,
        "enable_registration": True,
        "enable_registration_captcha": False,
        "macaroon_secret_key": "not even a little secret",
        "password_providers": [],
        "worker_app": None,
        "block_non_admin_invites": False,
        "federation_domain_whitelist": None,
        "filter_timeline_limit": 5000,
        "user_directory_search_all_users": False,
        "user_consent_server_notice_content": None,
        "block_events_without_consent_error": None,
        "user_consent_at_registration": False,
        "user_consent_policy_name": "Privacy Policy",
        "media_storage_providers": [],
        "autocreate_auto_join_rooms": True,
        "auto_join_rooms": [],
        "limit_usage_by_mau": False,
        "hs_disabled": False,
        "hs_disabled_message": "",
        "max_mau_value": 50,
        "mau_trial_days": 0,
        "mau_stats_only": False,
        "mau_limits_reserved_threepids": [],
        "admin_contact": None,
        "rc_message": {"per_second": 10000, "burst_count": 10000},
        "rc_registration": {"per_second": 10000, "burst_count": 10000},
        "rc_login": {
            "address": {"per_second": 10000, "burst_count": 10000},
            "account": {"per_second": 10000, "burst_count": 10000},
            "failed_attempts": {"per_second": 10000, "burst_count": 10000},
        },
        "rc_joins": {
            "local": {"per_second": 10000, "burst_count": 10000},
            "remote": {"per_second": 10000, "burst_count": 10000},
        },
        "rc_joins_per_room": {"per_second": 10000, "burst_count": 10000},
        "rc_invites": {
            "per_room": {"per_second": 10000, "burst_count": 10000},
            "per_user": {"per_second": 10000, "burst_count": 10000},
        },
        "rc_3pid_validation": {"per_second": 10000, "burst_count": 10000},
        "saml2_enabled": False,
        "public_baseurl": None,
        "default_identity_server": None,
        "key_refresh_interval": 24 * 60 * 60 * 1000,
        "old_signing_keys": {},
        "tls_fingerprints": [],
        "use_frozen_dicts": False,
        # We need a sane default_room_version, otherwise attempts to create
        # rooms will fail.
        "default_room_version": DEFAULT_ROOM_VERSION,
        # disable user directory updates, because they get done in the
        # background, which upsets the test runner. Setting this to an
        # (obviously) fake worker name disables updating the user directory.
        "update_user_directory_from_worker": "does_not_exist_worker_name",
        "caches": {"global_factor": 1, "sync_response_cache_duration": 0},
        "listeners": [{"port": 0, "type": "http"}],
    }

    if parse:
        config = HomeServerConfig()
        config.parse_config_dict(config_dict, "", "")
        return config

    return config_dict


def mock_getRawHeaders(headers=None):  # type: ignore[no-untyped-def]
    headers = headers if headers is not None else {}

    def getRawHeaders(name, default=None):  # type: ignore[no-untyped-def]
        # If the requested header is present, the real twisted function returns
        # List[str] if name is a str and List[bytes] if name is a bytes.
        # This mock doesn't support that behaviour.
        # Fortunately, none of the current callers of mock_getRawHeaders() provide a
        # headers dict, so we don't encounter this discrepancy in practice.
        return headers.get(name, default)

    return getRawHeaders


T = TypeVar("T")


def checked_cast(type_: type[T], x: object) -> T:
    """A version of typing.cast that is checked at runtime.

    We have our own function for this for two reasons:

    1. typing.cast itself is deliberately a no-op at runtime, see
       https://docs.python.org/3/library/typing.html#typing.cast
    2. To help workaround a mypy-zope bug https://github.com/Shoobx/mypy-zope/issues/91
       where mypy would erroneously consider `isinstance(x, type)` to be false in all
       circumstances.

    For this to make sense, `T` needs to be something that `isinstance` can check; see
        https://docs.python.org/3/library/functions.html?highlight=isinstance#isinstance
        https://docs.python.org/3/glossary.html#term-abstract-base-class
        https://docs.python.org/3/library/typing.html#typing.runtime_checkable
    for more details.
    """
    assert isinstance(x, type_)
    return x


def setup_awaitable_errors() -> Callable[[], None]:
    """
    Convert warnings from a non-awaited coroutines into errors.
    """
    warnings.simplefilter("error", RuntimeWarning)

    # State shared between unraisablehook and check_for_unraisable_exceptions.
    unraisable_exceptions = []
    orig_unraisablehook = sys.unraisablehook

    def unraisablehook(unraisable: "UnraisableHookArgs") -> None:
        unraisable_exceptions.append(unraisable.exc_value)

    def cleanup() -> None:
        """
        A method to be used as a clean-up that fails a test-case if there are any new unraisable exceptions.
        """
        sys.unraisablehook = orig_unraisablehook
        if unraisable_exceptions:
            exc = unraisable_exceptions.pop()
            assert exc is not None
            raise exc

    sys.unraisablehook = unraisablehook

    return cleanup



# Copyright (C) 2020, 2024 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# from jwcrypto import jwe, jwk, jwt
import logging
from http import HTTPStatus
from typing import TYPE_CHECKING, Any, ClassVar

from synapse.rest import admin
from synapse.rest.client import (
    account_data,
    login,
    notifications,
    presence,
    profile,
    room,
    room_upgrade_rest_servlet,
)
from synapse.server import HomeServer
from synapse.types import UserID, create_requester
from synapse.util.clock import Clock
from twisted.internet.testing import MemoryReactor

import tests.utils.homeserver_testcase as synapsetest
from famedly_control_synapse.rest.types import CreateManagedRoomRequest

logger = logging.getLogger(__name__)
# ruff: noqa: E501
# We don't care about long lines in our testdata

if TYPE_CHECKING:
    from synapse.storage.databases.main import DataStore
    from synapse.storage.databases.main.room import RoomWorkerStore


class ModuleApiTestCase(synapsetest.HomeserverTestCase):
    server_name_for_this_server = "testserver.com"
    _room_counter = 0
    BASE_PATH = "/_famedlyControl/v1/managedRooms"
    LIST_PATH = BASE_PATH + "/rooms"
    CREATE_PATH = BASE_PATH + "/createRoom"

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    servlets: ClassVar[list] = [
        admin.register_servlets,
        account_data.register_servlets,
        login.register_servlets,
        room.register_servlets,
        room_upgrade_rest_servlet.register_servlets,
        presence.register_servlets,
        profile.register_servlets,
        notifications.register_servlets,
    ]

    def prepare(
        self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer
    ) -> None:
        super().prepare(reactor, clock, homeserver)
        self.store: DataStore | RoomWorkerStore = homeserver.get_datastores().main
        self.storage_controllers = homeserver.get_storage_controllers()
        self.module_api = homeserver.get_module_api()
        self.event_creation_handler = homeserver.get_event_creation_handler()
        self.auth_handler = homeserver.get_auth_handler()
        self.creator = self.register_user("room_creator", "password", admin=True)
        self.creator_access_token = self.login("room_creator", "password")
        self.creator_token_id = self.get_success(
            self.hs.get_datastores().main.add_access_token_to_user(
                self.creator,
                "dummy",
                None,
                None,
            )
        )
        self.requester = create_requester(
            user_id=UserID.from_string(self.creator),
            access_token_id=self.creator_token_id,
        )
        self.invitee = self.register_user("invitee", "password")

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()
        if "modules" not in conf:
            conf["modules"] = [
                {
                    "module": "famedly_control_synapse.FamedlyControl",
                    "config": {
                        "famedly_control": {
                            "api_url": "http://dummy.test/famedlyControl",
                            "access_token": "dummy_token_for_testing",
                        },
                        "auth_provider": "https://idp.example.com/",
                        "sync_enabled": False,
                    },
                }
            ]
        return conf

    def _test_get_membership(
        self, room: str, members: list[str], expect_code: int = 200
    ) -> None:
        """Helper method to check the membership of a room.
        Returns 200 if the user is a member. If not, returns 403 accordingly."""
        for member in members:
            path = "/rooms/%s/state/m.room.member/%s" % (room, member)
            channel = self.make_request(
                "GET", path, access_token=self.creator_access_token
            )
            self.assertEqual(expect_code, channel.code)

    def _create_managed_room(
        self, name: str = "Test Room", groups: list[str] | None = None
    ) -> str:
        """Helper method to create a managed room with groups.
        At the moment requires mock for the get_group_members
        Returns room ID of the created room."""
        self._room_counter += 1
        config = CreateManagedRoomRequest(
            room_alias_name=f"test_room_{self._room_counter}",
            name=name,
            room_version="12",
            topic=f"Topic for {name}",
            groups=["test_group"],
        )
        if groups:
            config.groups = groups
        channel = self.make_request(
            method="POST",
            path=self.CREATE_PATH,
            content=config.model_dump(),
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result
        return channel.json_body["room_id"]



# Copyright 2018-2021 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import hashlib
import ipaddress
import json
import logging
import os
import os.path
import uuid
from collections import deque
from collections.abc import Awaitable, Callable, Iterable, MutableMapping, Sequence
from io import SEEK_END, BytesIO
from typing import Any, TypeVar, Union, cast
from unittest.mock import Mock

import attr
import twisted
from incremental import Version
from synapse.api.constants import MAX_REQUEST_SIZE
from synapse.config.database import DatabaseConnectionConfig
from synapse.config.homeserver import HomeServerConfig
from synapse.events.presence_router import load_legacy_presence_router
from synapse.handlers.auth import load_legacy_password_auth_providers
from synapse.http.site import SynapseRequest
from synapse.logging.context import ContextResourceUsage
from synapse.module_api.callbacks.spamchecker_callbacks import load_legacy_spam_checkers
from synapse.module_api.callbacks.third_party_event_rules_callbacks import (
    load_legacy_third_party_event_rules,
)
from synapse.server import HomeServer
from synapse.storage import DataStore
from synapse.storage.database import LoggingDatabaseConnection, make_pool
from synapse.storage.engines import BaseDatabaseEngine, create_engine
from synapse.types import ISynapseReactor, JsonDict
from synapse.util.clock import Clock
from twisted.enterprise import adbapi
from twisted.internet import address, tcp, threads, udp
from twisted.internet._resolver import SimpleResolverComplexifier
from twisted.internet.defer import Deferred, fail, maybeDeferred, succeed
from twisted.internet.error import DNSLookupError
from twisted.internet.interfaces import (
    IAddress,
    IConnector,
    IConsumer,
    IHostnameResolver,
    IListeningPort,
    IProducer,
    IProtocol,
    IPullProducer,
    IPushProducer,
    IReactorPluggableNameResolver,
    IReactorTime,
    IResolverSimple,
    ITransport,
)
from twisted.internet.protocol import ClientFactory, DatagramProtocol, Factory
from twisted.internet.testing import AccumulatingProtocol, MemoryReactorClock
from twisted.python import threadpool
from twisted.python.failure import Failure
from twisted.web.http_headers import Headers
from twisted.web.resource import IResource
from twisted.web.server import Request, Site
from typing_extensions import ParamSpec
from zope.interface import implementer

from tests.utils.config import (
    POSTGRES_BASE_DB,
    POSTGRES_HOST,
    POSTGRES_PASSWORD,
    POSTGRES_PORT,
    POSTGRES_USER,
    SQLITE_PERSIST_DB,
    USE_POSTGRES_FOR_TESTS,
    default_config,
)

logger = logging.getLogger(__name__)

R = TypeVar("R")
P = ParamSpec("P")

# the type of thing that can be passed into `make_request` in the headers list
CustomHeaderType = tuple[str | bytes, str | bytes]

# A pre-prepared SQLite DB that is used as a template when creating new SQLite
# DB each test run. This dramatically speeds up test set up when using SQLite.
PREPPED_SQLITE_DB_CONN: LoggingDatabaseConnection | None = None


class TimedOutError(Exception):
    """
    A web query timed out.
    """


@implementer(ITransport, IPushProducer, IConsumer)
@attr.s(auto_attribs=True)
class FakeChannel:
    """
    A fake Twisted Web Channel (the part that interfaces with the
    wire).

    See twisted.web.http.HTTPChannel.
    """

    site: Union[Site, "FakeSite"]
    _reactor: MemoryReactorClock
    result: dict = attr.Factory(dict)
    _ip: str = "127.0.0.1"
    _producer: IPullProducer | IPushProducer | None = None
    resource_usage: ContextResourceUsage | None = None
    _request: Request | None = None

    @property
    def request(self) -> Request:
        assert self._request is not None
        return self._request

    @request.setter
    def request(self, request: Request) -> None:
        assert self._request is None
        self._request = request

    @property
    def json_body(self) -> JsonDict:
        body = json.loads(self.text_body)
        assert isinstance(body, dict)
        return body

    @property
    def json_list(self) -> list[JsonDict]:
        body = json.loads(self.text_body)
        assert isinstance(body, list)
        return body

    @property
    def text_body(self) -> str:
        """The body of the result, utf-8-decoded.

        Raises an exception if the request has not yet completed.
        """
        if not self.is_finished():
            msg = "Request not yet completed"
            raise Exception(msg)
        return self.result["body"].decode("utf8")

    def is_finished(self) -> bool:
        """check if the response has been completely received"""
        return self.result.get("done", False)

    @property
    def code(self) -> int:
        if not self.result:
            msg = "No result yet."
            raise Exception(msg)
        return int(self.result["code"])

    @property
    def headers(self) -> Headers:
        if not self.result:
            msg = "No result yet."
            raise Exception(msg)
        h = Headers()
        for i in self.result["headers"]:
            h.addRawHeader(*i)
        return h

    def writeHeaders(
        self, version: bytes, code: bytes, reason: bytes, headers: Headers
    ) -> None:
        self.result["version"] = version
        self.result["code"] = code
        self.result["reason"] = reason
        self.result["headers"] = headers

    def write(self, data: bytes) -> None:
        assert isinstance(data, bytes), "Should be bytes! " + repr(data)

        if "body" not in self.result:
            self.result["body"] = b""

        self.result["body"] += data

    def writeSequence(self, data: Iterable[bytes]) -> None:
        for x in data:
            self.write(x)

    def loseConnection(self) -> None:
        self.unregisterProducer()
        self.transport.loseConnection()

    # Type ignore: mypy doesn't like the fact that producer isn't an IProducer.
    def registerProducer(self, producer: IProducer, streaming: bool) -> None:
        # TODO: This should ensure that the IProducer is an IPushProducer or
        # IPullProducer, unfortunately twisted.protocols.basic.FileSender does
        # implement those, but doesn't declare it.
        self._producer = cast(IPushProducer | IPullProducer, producer)
        self.producerStreaming = streaming

        def _produce() -> None:
            if self._producer:
                self._producer.resumeProducing()
                self._reactor.callLater(0.1, _produce)

        if not streaming:
            self._reactor.callLater(0.0, _produce)

    def unregisterProducer(self) -> None:
        if self._producer is None:
            return

        self._producer = None

    def stopProducing(self) -> None:
        if self._producer is not None:
            self._producer.stopProducing()

    def pauseProducing(self) -> None:
        raise NotImplementedError

    def resumeProducing(self) -> None:
        raise NotImplementedError

    def requestDone(self, _self: Request) -> None:
        self.result["done"] = True
        if isinstance(_self, SynapseRequest):
            assert _self.logcontext is not None
            self.resource_usage = _self.logcontext.get_resource_usage()

    def getPeer(self) -> IAddress:
        # We give an address so that getClientAddress/getClientIP returns a non null entry,
        # causing us to record the MAU
        return cast(IAddress, address.IPv4Address("TCP", self._ip, 3423))

    def getHost(self) -> IAddress:
        # this is called by Request.__init__ to configure Request.host.
        return cast(IAddress, address.IPv4Address("TCP", "127.0.0.1", 8888))

    def isSecure(self) -> bool:
        return False

    @property
    def transport(self) -> "FakeChannel":
        return self

    def await_result(self, timeout_ms: int = 1000) -> None:
        """
        Wait until the request is finished.
        """
        end_time = self._reactor.seconds() + timeout_ms / 1000.0
        self._reactor.run()

        while not self.is_finished():
            # If there's a producer, tell it to resume producing so we get content
            if self._producer:
                self._producer.resumeProducing()

            if self._reactor.seconds() > end_time:
                msg = "Timed out waiting for request to finish."
                raise TimedOutError(msg)

            self._reactor.advance(0.1)

    def extract_cookies(self, cookies: MutableMapping[str, str]) -> None:
        """Process the contents of any Set-Cookie headers in the response

        Any cookies found are added to the given dict
        """
        headers = self.headers.getRawHeaders("Set-Cookie")
        if not headers:
            return

        for h in headers:
            parts = h.split(";")
            k, v = parts[0].split("=", maxsplit=1)
            cookies[k] = v


class FakeSite:
    """
    A fake Twisted Web Site, with mocks of the extra things that
    Synapse adds.
    """

    server_version_string = b"1"
    site_tag = "test"
    access_logger = logging.getLogger("synapse.access.http.fake")

    def __init__(
        self,
        resource: IResource,
        reactor: IReactorTime,
        experimental_cors_msc3886: bool = False,
    ):
        """

        Args:
            resource: the resource to be used for rendering all requests
        """
        self._resource = resource
        self.reactor = reactor
        self.experimental_cors_msc3886 = experimental_cors_msc3886

    def getResourceFor(self, _request: Request) -> IResource:
        return self._resource


def make_request(
    reactor: MemoryReactorClock,
    site: Site | FakeSite,
    method: bytes | str,
    path: bytes | str,
    content: bytes | str | JsonDict = b"",
    access_token: str | None = None,
    request: type[Request] = SynapseRequest,
    shorthand: bool = True,
    federation_auth_origin: bytes | None = None,
    content_is_form: bool = False,
    await_result: bool = True,
    custom_headers: Iterable[CustomHeaderType] | None = None,
    client_ip: str = "127.0.0.1",
) -> FakeChannel:
    """
    Make a web request using the given method, path and content, and render it

    Returns the fake Channel object which records the response to the request.

    Args:
        reactor:
        site: The twisted Site to use to render the request
        method: The HTTP request method ("verb").
        path: The HTTP path, suitably URL encoded (e.g. escaped UTF-8 & spaces and such).
        content: The body of the request. JSON-encoded, if a str of bytes.
        access_token: The access token to add as authorization for the request.
        request: The request class to create.
        shorthand: Whether to try and be helpful and prefix the given URL
            with the usual REST API path, if it doesn't contain it.
        federation_auth_origin: if set to not-None, we will add a fake
            Authorization header pretending to be the given server name.
        content_is_form: Whether the content is URL encoded form data. Adds the
            'Content-Type': 'application/x-www-form-urlencoded' header.
        await_result: whether to wait for the request to complete rendering. If true,
             will pump the reactor until the the renderer tells the channel the request
             is finished.
        custom_headers: (name, value) pairs to add as request headers
        client_ip: The IP to use as the requesting IP. Useful for testing
            ratelimiting.

    Returns:
        channel
    """
    if not isinstance(method, bytes):
        method = method.encode("ascii")

    if not isinstance(path, bytes):
        path = path.encode("ascii")

    # Decorate it to be the full path, if we're using shorthand
    if (
        shorthand
        and not path.startswith(b"/_matrix")
        and not path.startswith(b"/_synapse")
    ):
        if path.startswith(b"/"):
            path = path[1:]
        path = b"/_matrix/client/r0/" + path

    if not path.startswith(b"/"):
        path = b"/" + path

    if isinstance(content, dict):
        content = json.dumps(content).encode("utf8")
    if isinstance(content, str):
        content = content.encode("utf8")

    channel = FakeChannel(site, reactor, ip=client_ip)  # type: ignore[call-arg]

    req = request(
        channel,
        site,
        our_server_name="test_server",
        max_request_body_size=MAX_REQUEST_SIZE,
    )
    channel.request = req

    req.content = BytesIO(content)  # type: ignore[assignment]
    # Twisted expects to be at the end of the content when parsing the request.
    req.content.seek(0, SEEK_END)  # type: ignore[attr-defined]

    # If `Content-Length` was passed in as a custom header, don't automatically add it
    # here.
    if custom_headers is None or not any(
        (k if isinstance(k, bytes) else k.encode("ascii")) == b"Content-Length"
        for k, _ in custom_headers
    ):
        # Old version of Twisted (<20.3.0) have issues with parsing x-www-form-urlencoded
        # bodies if the Content-Length header is missing
        req.requestHeaders.addRawHeader(
            b"Content-Length", str(len(content)).encode("ascii")
        )

    if access_token:
        req.requestHeaders.addRawHeader(
            b"Authorization", b"Bearer " + access_token.encode("ascii")
        )

    if federation_auth_origin is not None:
        req.requestHeaders.addRawHeader(
            b"Authorization",
            b"X-Matrix origin=%s,key=,sig=" % (federation_auth_origin,),
        )

    if content:
        if content_is_form:
            req.requestHeaders.addRawHeader(
                b"Content-Type", b"application/x-www-form-urlencoded"
            )
        else:
            # Assume the body is JSON
            req.requestHeaders.addRawHeader(b"Content-Type", b"application/json")

    if custom_headers:
        for k, v in custom_headers:
            req.requestHeaders.addRawHeader(k, v)

    req.parseCookies()
    req.requestReceived(method, path, b"1.1")

    if await_result:
        channel.await_result()

    return channel


# ISynapseReactor implies IReactorPluggableNameResolver, but explicitly
# marking this as an implementer of the latter seems to keep mypy-zope happier.
@implementer(IReactorPluggableNameResolver, ISynapseReactor)
class ThreadedMemoryReactorClock(MemoryReactorClock):
    """
    A MemoryReactorClock that supports callFromThread.
    """

    def __init__(self) -> None:
        self.threadpool = ThreadPool(self)

        self._tcp_callbacks: dict[tuple[str, int], Callable] = {}
        self._udp: list[udp.Port] = []
        self.lookups: dict[str, str] = {}
        self._thread_callbacks: deque[Callable[..., R]] = deque()

        lookups = self.lookups

        @implementer(IResolverSimple)
        class FakeResolver:
            def getHostByName(
                self, name: str, _timeout: Sequence[int] | None = None
            ) -> "Deferred[str]":
                if name not in lookups:
                    return fail(DNSLookupError(f"OH NO: unknown {name}"))
                return succeed(lookups[name])

        # In order for the TLS protocol tests to work, modify _get_default_clock
        # on newer Twisted versions to use the test reactor's clock.
        #
        # This is *super* dirty since it is never undone and relies on the next
        # test to overwrite it.
        if twisted.version > Version("Twisted", 23, 8, 0):
            from twisted.protocols import tls

            tls._get_default_clock = lambda: self

        self.nameResolver = SimpleResolverComplexifier(FakeResolver())
        super().__init__()

    def installNameResolver(self, resolver: IHostnameResolver) -> IHostnameResolver:
        raise NotImplementedError

    def listenUDP(
        self,
        port: int,
        protocol: DatagramProtocol,
        interface: str = "",
        maxPacketSize: int = 8196,
    ) -> udp.Port:
        p = udp.Port(port, protocol, interface, maxPacketSize, self)
        p.startListening()
        self._udp.append(p)
        return p

    def callFromThread(
        self,
        # noqa: A002, `callable` is inherited from IReactorFromThreads.callFromThread
        callable: Callable[..., Any],
        *args: object,
        **kwargs: object,
    ) -> None:
        """
        Make the callback fire in the next reactor iteration.
        """

        def cb():
            return callable(*args, **kwargs)

        # it's not safe to call callLater() here, so we append the callback to a
        # separate queue.
        self._thread_callbacks.append(cb)

    def callInThread(
        self,
        # noqa: A002, `callable` is inherited from IReactorFromThreads.callInThread
        callable: Callable[..., Any],
        *args: object,
        **kwargs: object,
    ) -> None:
        raise NotImplementedError

    def suggestThreadPoolSize(self, size: int) -> None:
        raise NotImplementedError

    def getThreadPool(self) -> "threadpool.ThreadPool":
        # Cast to match super-class.
        return cast(threadpool.ThreadPool, self.threadpool)

    def add_tcp_client_callback(
        self, host: str, port: int, callback: Callable[[], None]
    ) -> None:
        """Add a callback that will be invoked when we receive a connection
        attempt to the given IP/port using `connectTCP`.

        Note that the callback gets run before we return the connection to the
        client, which means callbacks cannot block while waiting for writes.
        """
        self._tcp_callbacks[(host, port)] = callback

    def connectUNIX(
        self,
        _address: str,
        _factory: ClientFactory,
        _timeout: float = 30,
        _checkPID: int = 0,
    ) -> IConnector:
        """
        Unix sockets aren't supported for unit tests yet. Make it obvious to any
        developer trying it out that they will need to do some work before being able
        to use it in tests.
        """
        msg = "Unix sockets are not implemented for tests yet, sorry."
        raise Exception(msg)

    def listenUNIX(
        self,
        _address: str,
        _factory: Factory,
        _backlog: int = 50,
        _mode: int = 0o666,
        _wantPID: int = 0,
    ) -> IListeningPort:
        """
        Unix sockets aren't supported for unit tests yet. Make it obvious to any
        developer trying it out that they will need to do some work before being able
        to use it in tests.
        """
        msg = "Unix sockets are not implemented for tests, sorry"
        raise Exception(msg)

    def connectTCP(
        self,
        host: str,
        port: int,
        factory: ClientFactory,
        timeout: float = 30,
        _bindAddress: tuple[str, int] | None = None,
    ) -> IConnector:
        """Fake L{IReactorTCP.connectTCP}."""

        conn = super().connectTCP(
            host, port, factory, timeout=timeout, bindAddress=None
        )
        if self.lookups and host in self.lookups:
            validate_connector(conn, self.lookups[host])

        callback = self._tcp_callbacks.get((host, port))
        if callback:
            callback()

        return conn

    def advance(self, amount: float) -> None:
        # first advance our reactor's time, and run any "callLater" callbacks that
        # makes ready
        super().advance(amount)

        # now run any "callFromThread" callbacks
        while True:
            try:
                callback = self._thread_callbacks.popleft()
            except IndexError:
                break
            callback()

            # check for more "callLater" callbacks added by the thread callback
            # This isn't required in a regular reactor, but it ends up meaning that
            # our database queries can complete in a single call to `advance` [1] which
            # simplifies tests.
            #
            # [1]: we replace the threadpool backing the db connection pool with a
            # mock ThreadPool which doesn't really use threads; but we still use
            # reactor.callFromThread to feed results back from the db functions to the
            # main thread.
            super().advance(0)


def validate_connector(connector: tcp.Connector, expected_ip: str) -> None:
    """Try to validate the obtained connector as it would happen when
    synapse is running and the connection will be established.

    This method will raise a useful exception when necessary, else it will
    just do nothing.

    This is in order to help catch quirks related to reactor.connectTCP,
    since when called directly, the connector's destination will be of type
    IPv4Address, with the hostname as the literal host that was given (which
    could be an IPv6-only host or an IPv6 literal).

    But when called from reactor.connectTCP *through* e.g. an Endpoint, the
    connector's destination will contain the specific IP address with the
    correct network stack class.

    Note that testing code paths that use connectTCP directly should not be
    affected by this check, unless they specifically add a test with a
    matching reactor.lookups[HOSTNAME] = "IPv6Literal", where reactor is of
    type ThreadedMemoryReactorClock.
    For an example of implementing such tests, see test/handlers/send_email.py.
    """
    destination = connector.getDestination()

    # We use address.IPv{4,6}Address to check what the reactor thinks it is
    # is sending but check for validity with ipaddress.IPv{4,6}Address
    # because they fail with IPs on the wrong network stack.
    cls_mapping = {
        address.IPv4Address: ipaddress.IPv4Address,
        address.IPv6Address: ipaddress.IPv6Address,
    }

    cls = cls_mapping.get(destination.__class__)

    if cls is not None:
        try:
            cls(expected_ip)
        except Exception as exc:
            msg = f"Invalid IP type and resolution for {destination}. Expected {expected_ip} to be {cls.__name__}"
            raise ValueError(msg) from exc
    else:
        msg = f"Unknown address type {destination.__class__.__name__} for {destination}"
        raise ValueError(msg)


def make_fake_db_pool(
    reactor: ISynapseReactor,
    db_config: DatabaseConnectionConfig,
    engine: BaseDatabaseEngine,
    server_name: str,
) -> adbapi.ConnectionPool:
    """Wrapper for `make_pool` which builds a pool which runs db queries synchronously.

    For more deterministic testing, we don't use a regular db connection pool: instead
    we run all db queries synchronously on the test reactor's main thread. This function
    is a drop-in replacement for the normal `make_pool` which builds such a connection
    pool.
    """
    pool = make_pool(
        reactor=reactor, db_config=db_config, engine=engine, server_name=server_name
    )

    def runWithConnection(
        func: Callable[..., R], *args: Any, **kwargs: Any
    ) -> Awaitable[R]:
        return threads.deferToThreadPool(
            pool._reactor,
            pool.threadpool,
            pool._runWithConnection,
            func,
            *args,
            **kwargs,
        )

    def runInteraction(
        desc: str, func: Callable[..., R], *args: Any, **kwargs: Any
    ) -> Awaitable[R]:
        return threads.deferToThreadPool(
            pool._reactor,
            pool.threadpool,
            pool._runInteraction,
            desc,
            func,
            *args,
            **kwargs,
        )

    pool.runWithConnection = runWithConnection  # type: ignore[method-assign]
    pool.runInteraction = runInteraction  # type: ignore[assignment]
    # Replace the thread pool with a threadless 'thread' pool
    pool.threadpool = ThreadPool(reactor)
    pool.running = True
    return pool


class ThreadPool:
    """
    Threadless thread pool.

    See twisted.python.threadpool.ThreadPool
    """

    def __init__(self, reactor: IReactorTime):
        self._reactor = reactor

    def start(self) -> None:
        pass

    def stop(self) -> None:
        pass

    def callInThreadWithCallback(
        self,
        onResult: Callable[[bool, Failure | R], None],
        function: Callable[P, R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> "Deferred[None]":
        def _(res: Any) -> None:
            if isinstance(res, Failure):
                onResult(False, res)
            else:
                onResult(True, res)

        d: Deferred[None] = Deferred()
        d.addCallback(lambda _: function(*args, **kwargs))
        d.addBoth(_)
        self._reactor.callLater(0, d.callback, True)
        return d


def _make_test_homeserver_synchronous(server: HomeServer) -> None:
    """
    Make the given test homeserver's database interactions synchronous.
    """

    clock = server.get_clock()

    for database in server.get_datastores().databases:
        pool = database._db_pool

        def wrap_pool(pool):
            def runWithConnection(
                func: Callable[..., R], *args: Any, **kwargs: Any
            ) -> Awaitable[R]:
                return threads.deferToThreadPool(
                    pool._reactor,
                    pool.threadpool,
                    pool._runWithConnection,
                    func,
                    *args,
                    **kwargs,
                )

            def runInteraction(
                desc: str, func: Callable[..., R], *args: Any, **kwargs: Any
            ) -> Awaitable[R]:
                return threads.deferToThreadPool(
                    pool._reactor,
                    pool.threadpool,
                    pool._runInteraction,
                    desc,
                    func,
                    *args,
                    **kwargs,
                )

            pool.runWithConnection = runWithConnection  # type: ignore[method-assign]
            pool.runInteraction = runInteraction  # type: ignore[assignment]
            # Replace the thread pool with a threadless 'thread' pool
            pool.threadpool = ThreadPool(clock._reactor)
            pool.running = True

        wrap_pool(pool)

    # We've just changed the Databases to run DB transactions on the same
    # thread, so we need to disable the dedicated thread behaviour.
    server.get_datastores().main.USE_DEDICATED_DB_THREADS_FOR_EVENT_FETCHING = False


def get_clock() -> tuple[ThreadedMemoryReactorClock, Clock]:
    reactor = ThreadedMemoryReactorClock()
    hs_clock = Clock(reactor, server_name="test_server")
    return reactor, hs_clock


@implementer(ITransport)
@attr.s(cmp=False, auto_attribs=True)
class FakeTransport:
    """
    A twisted.internet.interfaces.ITransport implementation which sends all its data
    straight into an IProtocol object: it exists to connect two IProtocols together.

    To use it, instantiate it with the receiving IProtocol, and then pass it to the
    sending IProtocol's makeConnection method:

        server = HTTPChannel()
        client.makeConnection(FakeTransport(server, self.reactor))

    If you want bidirectional communication, you'll need two instances.
    """

    other: IProtocol
    """The Protocol object which will receive any data written to this transport.
    """

    _reactor: IReactorTime
    """Test reactor
    """

    _protocol: "IProtocol | None" = None
    """The Protocol which is producing data for this transport. Optional, but if set
    will get called back for connectionLost() notifications etc.
    """

    _peer_address: IAddress = attr.Factory(
        lambda: address.IPv4Address("TCP", "127.0.0.1", 5678)
    )
    """The value to be returned by getPeer"""

    _host_address: IAddress = attr.Factory(
        lambda: address.IPv4Address("TCP", "127.0.0.1", 1234)
    )
    """The value to be returned by getHost"""

    disconnecting = False
    disconnected = False
    connected = True
    buffer: bytes = b""
    producer: IPushProducer | None = None
    autoflush: bool = True

    def getPeer(self) -> IAddress:
        return self._peer_address

    def getHost(self) -> IAddress:
        return self._host_address

    def loseConnection(self) -> None:
        if not self.disconnecting:
            logger.info("FakeTransport: loseConnection()")
            self.disconnecting = True
            if self._protocol:
                self._protocol.connectionLost(
                    Failure(RuntimeError("FakeTransport.loseConnection()"))
                )

            # if we still have data to write, delay until that is done
            if self.buffer:
                logger.info(
                    "FakeTransport: Delaying disconnect until buffer is flushed"
                )
            else:
                self.connected = False
                self.disconnected = True

    def abortConnection(self) -> None:
        logger.info("FakeTransport: abortConnection()")

        if not self.disconnecting:
            self.disconnecting = True
            if self._protocol:
                self._protocol.connectionLost(None)  # type: ignore[arg-type]

        self.disconnected = True

    def pauseProducing(self) -> None:
        if not self.producer:
            return

        self.producer.pauseProducing()

    def resumeProducing(self) -> None:
        if not self.producer:
            return
        self.producer.resumeProducing()

    def unregisterProducer(self) -> None:
        if not self.producer:
            return

        self.producer = None

    def registerProducer(self, producer: IPushProducer, streaming: bool) -> None:
        self.producer = producer
        self.producerStreaming = streaming

        def _produce() -> None:
            if not self.producer:
                # we've been unregistered
                return
            # some implementations of IProducer (for example, FileSender)
            # don't return a deferred.
            d = maybeDeferred(self.producer.resumeProducing)
            d.addCallback(lambda _: self._reactor.callLater(0.1, _produce))

        if not streaming:
            self._reactor.callLater(0.0, _produce)

    def write(self, byt: bytes) -> None:
        if self.disconnecting:
            msg = "Writing to disconnecting FakeTransport"
            raise Exception(msg)

        self.buffer = self.buffer + byt

        # always actually do the write asynchronously. Some protocols (notably the
        # TLSMemoryBIOProtocol) get very confused if a read comes back while they are
        # still doing a write. Doing a callLater here breaks the cycle.
        if self.autoflush:
            self._reactor.callLater(0.0, self.flush)

    def writeSequence(self, seq: Iterable[bytes]) -> None:
        for x in seq:
            self.write(x)

    def flush(self, maxbytes: int | None = None) -> None:
        if not self.buffer:
            # nothing to do. Don't write empty buffers: it upsets the
            # TLSMemoryBIOProtocol
            return

        if self.disconnected:
            return

        to_write = self.buffer[:maxbytes] if maxbytes is not None else self.buffer

        logger.info("%s->%s: %s", self._protocol, self.other, to_write)

        try:
            self.other.dataReceived(to_write)
        except Exception:
            logger.exception("Exception writing to protocol")
            return

        self.buffer = self.buffer[len(to_write) :]
        if self.buffer and self.autoflush:
            self._reactor.callLater(0.0, self.flush)

        if not self.buffer and self.disconnecting:
            logger.info("FakeTransport: Buffer now empty, completing disconnect")
            self.disconnected = True


def connect_client(
    reactor: ThreadedMemoryReactorClock, client_id: int
) -> tuple[IProtocol, AccumulatingProtocol]:
    """
    Connect a client to a fake TCP transport.

    Args:
        reactor
        factory: The connecting factory to build.
    """
    factory = reactor.tcpClients.pop(client_id)[2]
    client = factory.buildProtocol(None)
    server = AccumulatingProtocol()
    server.makeConnection(FakeTransport(client, reactor))  # type: ignore[call-arg]
    client.makeConnection(FakeTransport(server, reactor))  # type: ignore[call-arg]

    return client, server


class TestHomeServer(HomeServer):
    DATASTORE_CLASS = DataStore  # type: ignore[assignment]


def setup_test_homeserver(
    name: str = "test",
    config: HomeServerConfig | None = None,
    reactor: ISynapseReactor | None = None,
    homeserver_to_use: type[HomeServer] = TestHomeServer,
    **kwargs: Any,
) -> HomeServer:
    """
    Setup a homeserver suitable for running tests against.  Keyword arguments
    are passed to the Homeserver constructor.

    If no datastore is supplied, one is created and given to the homeserver.

    Calling this method directly is deprecated: you should instead derive from
    HomeserverTestCase.
    """
    if reactor is None:
        reactor = ThreadedMemoryReactorClock()

    if config is None:
        config = default_config(name, parse=True)

    config.caches.resize_all_caches()

    if "clock" not in kwargs:
        kwargs["clock"] = Clock(reactor, server_name=name)

    if USE_POSTGRES_FOR_TESTS:
        test_db = f"synapse_test_{uuid.uuid4().hex}"

        database_config = {
            "name": "psycopg2",
            "args": {
                "dbname": test_db,
                "host": POSTGRES_HOST,
                "password": POSTGRES_PASSWORD,
                "user": POSTGRES_USER,
                "port": POSTGRES_PORT,
                "cp_min": 1,
                "cp_max": 5,
            },
        }
    else:
        if SQLITE_PERSIST_DB:
            # The current working directory is in _trial_temp, so this gets created within that directory.
            test_db_location = os.path.abspath("test.db")
            logger.debug("Will persist db to %s", test_db_location)
            # Ensure each test gets a clean database.
            try:
                os.remove(test_db_location)
            except FileNotFoundError:
                pass
            else:
                logger.debug("Removed existing DB at %s", test_db_location)
        else:
            test_db_location = ":memory:"

        database_config = {
            "name": "sqlite3",
            "args": {"database": test_db_location, "cp_min": 1, "cp_max": 1},
        }

    if "db_txn_limit" in kwargs:
        database_config["txn_limit"] = kwargs["db_txn_limit"]

    database = DatabaseConnectionConfig("master", database_config)
    config.database.databases = [database]

    db_engine = create_engine(database.config)

    # Create the database before we actually try and connect to it, based off
    # the template database we generate in setupdb()
    if USE_POSTGRES_FOR_TESTS:
        db_conn = db_engine.module.connect(
            dbname=POSTGRES_BASE_DB,
            user=POSTGRES_USER,
            host=POSTGRES_HOST,
            port=POSTGRES_PORT,
            password=POSTGRES_PASSWORD,
        )
        db_engine.attempt_to_set_autocommit(db_conn, True)
        cur = db_conn.cursor()
        cur.execute(f"DROP DATABASE IF EXISTS {test_db};")
        cur.execute(f"CREATE DATABASE {test_db} WITH TEMPLATE {POSTGRES_BASE_DB};")
        cur.close()
        db_conn.close()

    hs = homeserver_to_use(
        name,
        config=config,
        reactor=reactor,
    )

    # Install @cache_in_self attributes
    for key, val in kwargs.items():
        setattr(hs, "_" + key, val)

    # Mock TLS
    hs.tls_server_context_factory = Mock()

    hs.setup()

    # bcrypt is far too slow to be doing in unit tests
    # Need to let the HS build an auth handler and then mess with it
    # because AuthHandler's constructor requires the HS, so we can't make one
    # beforehand and pass it in to the HS's constructor (chicken / egg)
    async def _hash(p: str) -> str:
        return hashlib.md5(p.encode("utf8")).hexdigest()  # noqa: S324

    hs.get_auth_handler().hash = _hash  # type: ignore[assignment]

    async def validate_hash(p: str, h: str) -> bool:
        return hashlib.md5(p.encode("utf8")).hexdigest() == h  # noqa: S324

    hs.get_auth_handler().validate_hash = validate_hash  # type: ignore[assignment]

    # Make the threadpool and database transactions synchronous for testing.
    _make_test_homeserver_synchronous(hs)

    # Load any configured modules into the homeserver
    module_api = hs.get_module_api()
    for module, module_config in hs.config.modules.loaded_modules:
        hs.room_control = module(config=module_config, api=module_api)  # type: ignore[attr-defined]
        logger.debug("Loaded module %s %r", module, module_config)

    load_legacy_spam_checkers(hs)
    load_legacy_third_party_event_rules(hs)
    load_legacy_presence_router(hs)
    load_legacy_password_auth_providers(hs)

    return hs



from typing import Literal

from pydantic import BaseModel, ConfigDict, Field
from synapse.api.constants import EventTypes, GuestAccess, Membership
from synapse.types import JsonDict

MANAGED_ROOM_TYPE = "de.famedly.managedRoom"
SYNC_TOKEN_TYPE = "de.famedly.roomControl.lastSyncToken.v1"


class CreationContent(BaseModel):
    """Pydantic model for CreationContent."""

    creator: str | None = None
    m_federate: Literal[False] = Field(
        default=False,
        alias="m.federate",
        description="Federation is disabled for managed rooms.",
    )
    predecessor: JsonDict | None = None
    room_version: str | None = (
        None  # The room version here is overridden. It is irrelevant which version comes here.
    )
    type: str | None = None

    model_config = ConfigDict(validate_by_name=True, validate_by_alias=True)


class PowerLevelEventContent(BaseModel):
    """Power level event content for overriding default power levels."""

    ban: int = 100
    events: dict[str, int] = Field(
        default_factory=lambda: {
            EventTypes.Name: 100,
            EventTypes.Topic: 100,
            EventTypes.PowerLevels: 100,
            EventTypes.JoinRules: 100,
            EventTypes.CanonicalAlias: 100,
            EventTypes.RoomAvatar: 100,
        }
    )
    events_default: int = 0
    invite: int = 100
    kick: int = 100
    redact: int = 100
    state_default: int = 100
    users: dict[str, int] = Field(default_factory=dict)
    users_default: int = 0
    notifications: dict[str, int] = Field(default_factory=dict)


class CreateManagedRoomRequest(BaseModel):
    """Request body for creating a managed room."""

    room_alias_name: str
    name: str
    topic: str | None = None
    creation_content: CreationContent = Field(default_factory=CreationContent)
    initial_state: list[JsonDict] = Field(
        default_factory=lambda: [
            {
                "type": EventTypes.GuestAccess,
                "state_key": "",
                "content": {"guest_access": GuestAccess.FORBIDDEN},
            },
            {
                "type": EventTypes.JoinRules,
                "state_key": "",
                "content": {"join_rule": Membership.INVITE},
            },
        ]
    )
    power_level_content_override: PowerLevelEventContent = Field(
        default_factory=PowerLevelEventContent
    )
    room_version: str | None = None
    groups: list[str]
    is_direct: Literal[False] = Field(
        default=False, description="is_direct cannot be enabled for managed rooms."
    )
    visibility: Literal["private"] = Field(
        default="private",
        description="Visibility is fixed to private for managed rooms.",
    )

    model_config = ConfigDict(extra="forbid")


class AssignGroupsToManagedRoomRequest(BaseModel):
    """Request body for assigning groups to a managed room."""

    groups: list[str]

    model_config = ConfigDict(extra="forbid")



# Copyright (C) 2026 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import logging
from typing import Any

from synapse.api.constants import EventTypes, Membership
from synapse.api.errors import Codes, SynapseError
from synapse.event_auth import get_user_power_level
from synapse.events import EventBase
from synapse.http.server import JsonResource
from synapse.module_api import ModuleApi
from synapse.types import StateMap

from famedly_control_synapse.client import FamedlyControlClient
from famedly_control_synapse.config import FamedlyControlConfig
from famedly_control_synapse.repository import ManagedRoomRepository
from famedly_control_synapse.rest.room import (
    MANAGED_ROOM_API_PREFIX,
    AssignGroupsToManagedRoomResource,
    CreateManagedRoomResource,
    ListManagedRoomsResource,
)
from famedly_control_synapse.room_handler import ManagedRoomHandler
from famedly_control_synapse.sync import GroupMembershipSyncer

logger = logging.getLogger(__name__)


class _SyncTriggerJsonResource(JsonResource):
    """JsonResource subclass that attempts to start the background sync
    on every incoming request. If the data isn't ready yet, the next API call
    will retry automatically.
    """

    def __init__(self, hs, syncer: GroupMembershipSyncer):
        super().__init__(hs)
        self._syncer = syncer

    def render(self, request):
        response_body = super().render(request)
        self._syncer.start()
        return response_body


class FamedlyControl:
    __version__ = "0.0.1"

    def __init__(self, config: FamedlyControlConfig, api: ModuleApi) -> None:
        self.api = api
        self.server_name = api.server_name
        self.clock = api._hs.get_clock()
        self.config = config
        self.client = FamedlyControlClient(self.api, config)
        self.room_handler = ManagedRoomHandler(self.api, self.config)
        self.repository = ManagedRoomRepository(api)

        if self.api.should_run_background_tasks():
            self.syncer = GroupMembershipSyncer(
                api, self.client, self.room_handler, self.repository, config
            )

            # Register servlets
            self.resource = _SyncTriggerJsonResource(self.api._hs, self.syncer)
            CreateManagedRoomResource(
                self.api, self.client, self.room_handler, self.repository
            ).register(self.resource)
            ListManagedRoomsResource(self.api, self.repository).register(self.resource)
            AssignGroupsToManagedRoomResource(
                self.api, self.client, self.room_handler, self.repository
            ).register(self.resource)
            self.api.register_web_resource(MANAGED_ROOM_API_PREFIX, self.resource)

            self.api._clock.call_when_running(self.syncer.start)

        self.api.register_third_party_rules_callbacks(
            check_event_allowed=self.check_event_allowed,
        )

        logger.info("Module initialized")

    @staticmethod
    def parse_config(config: dict[str, Any]) -> FamedlyControlConfig:
        return FamedlyControlConfig.model_validate(config)

    async def check_event_allowed(
        self, event: EventBase, state_events: StateMap[EventBase]
    ) -> tuple[bool, dict | None]:
        """Third-party rules callback that enforces membership and power level
        restrictions for managed rooms.
        """
        if event.type not in (EventTypes.Member, EventTypes.PowerLevels):
            return True, None

        create_event = state_events.get((EventTypes.Create, ""))
        if create_event is None:
            raise SynapseError(
                500,
                "Managed room is missing m.room.create state",
                Codes.UNKNOWN,
            )

        if not await self.repository.is_managed_room(
            event.room_id, create_event.sender
        ):
            return True, None

        if event.type == EventTypes.Member:
            return await self._check_membership_allowed(event, create_event.sender)
        elif event.type == EventTypes.PowerLevels:
            return await self._check_power_levels_allowed(
                event, state_events, create_event.sender
            )
        else:
            # cannot happen but makes the linter happy
            return True, None

    async def _check_membership_allowed(
        self, event: EventBase, admin_user: str
    ) -> tuple[bool, dict | None]:
        """Block membership changes in managed rooms unless sent by the admin.

        Joins are allowed because managed rooms are invite-only and only
        the admin can send invites, so a join is implicitly admin-authorised.
        """
        if event.sender == admin_user:
            return True, None

        membership = event.content.get("membership")
        if membership == Membership.JOIN:
            return True, None

        raise SynapseError(
            403,
            "Only the admin can edit membership in a managed room",
            Codes.FORBIDDEN,
        )

    async def _check_power_levels_allowed(
        self, event: EventBase, state_events: StateMap[EventBase], admin_user: str
    ) -> tuple[bool, dict | None]:
        """Block invalid power level changes in managed rooms.

        Enforced constraints:
        * The admin must remain the highest power level in the room.
        * No non-admin user may have a PL >= the admin's PL.
        * ``users_default`` must be strictly below the admin's PL.
        * Non-admin users must stay below sensitive event thresholds
          (``state_default``, ``m.room.power_levels``).
        * Membership actions (ban, kick, invite) must not exceed the
          admin's PL and must be strictly above all non-admin users.
        """
        new_content = event.content
        users = new_content.get("users", {})
        users_default = new_content.get("users_default", 0)
        # the admin isn't necessarily present in the users list for power level updates
        # this is the case for rooms >v12
        current_admin_pl = get_user_power_level(admin_user, state_events)
        admin_pl = users.get(admin_user, current_admin_pl)

        for user_id, pl in users.items():
            if user_id != admin_user and pl >= admin_pl:
                raise SynapseError(
                    403,
                    "No user can have a power level equal to or higher than the admin in a managed room",
                    Codes.FORBIDDEN,
                )

        if users_default >= admin_pl:
            raise SynapseError(
                403,
                "users_default cannot be equal to or higher than the admin power level in a managed room",
                Codes.FORBIDDEN,
            )

        # Prevent non-admin users from reaching sensitive event thresholds.
        state_default = new_content.get("state_default", 50)
        events = new_content.get("events", {})
        power_levels_pl = events.get(EventTypes.PowerLevels, state_default)
        sensitive_threshold = min(state_default, power_levels_pl)

        max_non_admin_pl = users_default

        for uid, pl in users.items():
            if uid != admin_user and pl > max_non_admin_pl:
                max_non_admin_pl = pl

        if users_default >= sensitive_threshold:
            raise SynapseError(
                403,
                "users_default cannot be equal to or above the threshold for sensitive state events in a managed room",
                Codes.FORBIDDEN,
            )

        for user_id, pl in users.items():
            if user_id != admin_user and pl >= sensitive_threshold:
                raise SynapseError(
                    403,
                    "non-admin user power level cannot be equal to or above the threshold for sensitive state events in a managed room",
                    Codes.FORBIDDEN,
                )

        # membership actions default to Matrix-spec values when not explicitly set.
        membership_action_defaults = {
            "ban": 50,
            "kick": 50,
            "invite": 50,
        }
        for action in ("ban", "kick", "invite"):
            action_pl = new_content.get(action)
            if action_pl is None:
                action_pl = membership_action_defaults[action]

            if action_pl > admin_pl:
                raise SynapseError(
                    403,
                    f"{action} power level cannot exceed admin power level in a managed room",
                    Codes.FORBIDDEN,
                )

            if action_pl <= max_non_admin_pl:
                raise SynapseError(
                    403,
                    f"{action} power level must be higher than all non-admin users in a managed room",
                    Codes.FORBIDDEN,
                )

        return True, None



import pytest
from pydantic import ValidationError

from famedly_control_synapse.rest.types import (
    CreateManagedRoomRequest,
    CreationContent,
    PowerLevelEventContent,
)


class TestCreationContent:
    def test_valid_creator_for_v10(self):
        content = CreationContent(creator="@alice:example.com")
        assert content.creator == "@alice:example.com"

    def test_m_federate_always_false(self):
        content = CreationContent()
        assert content.m_federate is False

        with pytest.raises(ValidationError):
            content = CreationContent(m_federate=True)


class TestPowerLevelEventContent:
    def test_valid_power_level_event_content(self):
        plc = PowerLevelEventContent(
            users={"@alice:example.com": 100, "@bob:example.com": 50}, users_default=0
        )
        assert plc.users["@alice:example.com"] == 100
        assert plc.users["@bob:example.com"] == 50
        assert plc.users_default == 0

    def test_invalid_users_type(self):
        with pytest.raises(ValidationError):
            PowerLevelEventContent(
                users=["@alice:example.com", "@bob:example.com"], users_default=0
            )


class TestCreateManagedRoomRequest:
    def test_create_managed_room_v10_request_valid(self):
        # Valid request
        req = CreateManagedRoomRequest(
            room_alias_name="testroom",
            name="Test Room",
            room_version="10",
            creation_content=CreationContent(creator="@alice:example.com"),
            power_level_content_override=PowerLevelEventContent(
                users={"@alice:example.com": 100}, users_default=0
            ),
            groups=["testgroup"],
        )
        assert req.room_alias_name == "testroom"
        assert req.creation_content.creator == "@alice:example.com"

    def test_create_managed_room_request_valid(self):
        # Valid request
        req = CreateManagedRoomRequest(
            room_alias_name="testroom",
            name="Test Room",
            room_version="12",
            creation_content=CreationContent(),
            power_level_content_override=PowerLevelEventContent(),
            groups=["testgroup"],
        )
        assert req.room_alias_name == "testroom"
        assert req.creation_content.creator is None
        assert req.initial_state == [
            {
                "content": {
                    "guest_access": "forbidden",
                },
                "state_key": "",
                "type": "m.room.guest_access",
            },
            {
                "content": {
                    "join_rule": "invite",
                },
                "state_key": "",
                "type": "m.room.join_rules",
            },
        ]
        assert req.is_direct is False
        assert req.visibility == "private"
